# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/agent.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='cogment/api/agent.proto',
  package='cogment',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x17\x63ogment/api/agent.proto\x12\x07\x63ogment\x1a\x18\x63ogment/api/common.proto\"z\n\x11\x41gentStartRequest\x12\x11\n\timpl_name\x18\x01 \x01(\t\x12$\n\x06\x63onfig\x18\x02 \x01(\x0b\x32\x14.cogment.ActorConfig\x12,\n\x0f\x61\x63tors_in_trial\x18\x03 \x03(\x0b\x32\x13.cogment.TrialActor\"\x11\n\x0f\x41gentStartReply\"D\n\x17\x41gentObservationRequest\x12)\n\x0bobservation\x18\x01 \x01(\x0b\x32\x14.cogment.Observation\"y\n\x10\x41gentActionReply\x12\x1f\n\x06\x61\x63tion\x18\x01 \x01(\x0b\x32\x0f.cogment.Action\x12 \n\x07rewards\x18\x02 \x03(\x0b\x32\x0f.cogment.Reward\x12\"\n\x08messages\x18\x03 \x03(\x0b\x32\x10.cogment.Message\"5\n\x12\x41gentRewardRequest\x12\x1f\n\x06reward\x18\x01 \x01(\x0b\x32\x0f.cogment.Reward\"\x12\n\x10\x41gentRewardReply\"9\n\x13\x41gentMessageRequest\x12\"\n\x08messages\x18\x01 \x03(\x0b\x32\x10.cogment.Message\"\x13\n\x11\x41gentMessageReply\"?\n\x0f\x41gentEndRequest\x12,\n\nfinal_data\x18\x01 \x01(\x0b\x32\x18.cogment.ActorPeriodData\"\x0f\n\rAgentEndReply2\xae\x03\n\rAgentEndpoint\x12\x41\n\x07OnStart\x12\x1a.cogment.AgentStartRequest\x1a\x18.cogment.AgentStartReply\"\x00\x12R\n\rOnObservation\x12 .cogment.AgentObservationRequest\x1a\x19.cogment.AgentActionReply\"\x00(\x01\x30\x01\x12\x44\n\x08OnReward\x12\x1b.cogment.AgentRewardRequest\x1a\x19.cogment.AgentRewardReply\"\x00\x12G\n\tOnMessage\x12\x1c.cogment.AgentMessageRequest\x1a\x1a.cogment.AgentMessageReply\"\x00\x12;\n\x05OnEnd\x12\x18.cogment.AgentEndRequest\x1a\x16.cogment.AgentEndReply\"\x00\x12:\n\x07Version\x12\x17.cogment.VersionRequest\x1a\x14.cogment.VersionInfo\"\x00\x62\x06proto3'
  ,
  dependencies=[cogment_dot_api_dot_common__pb2.DESCRIPTOR,])




_AGENTSTARTREQUEST = _descriptor.Descriptor(
  name='AgentStartRequest',
  full_name='cogment.AgentStartRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='impl_name', full_name='cogment.AgentStartRequest.impl_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='config', full_name='cogment.AgentStartRequest.config', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actors_in_trial', full_name='cogment.AgentStartRequest.actors_in_trial', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=62,
  serialized_end=184,
)


_AGENTSTARTREPLY = _descriptor.Descriptor(
  name='AgentStartReply',
  full_name='cogment.AgentStartReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=186,
  serialized_end=203,
)


_AGENTOBSERVATIONREQUEST = _descriptor.Descriptor(
  name='AgentObservationRequest',
  full_name='cogment.AgentObservationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='observation', full_name='cogment.AgentObservationRequest.observation', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=205,
  serialized_end=273,
)


_AGENTACTIONREPLY = _descriptor.Descriptor(
  name='AgentActionReply',
  full_name='cogment.AgentActionReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='action', full_name='cogment.AgentActionReply.action', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rewards', full_name='cogment.AgentActionReply.rewards', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='messages', full_name='cogment.AgentActionReply.messages', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=275,
  serialized_end=396,
)


_AGENTREWARDREQUEST = _descriptor.Descriptor(
  name='AgentRewardRequest',
  full_name='cogment.AgentRewardRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='reward', full_name='cogment.AgentRewardRequest.reward', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=398,
  serialized_end=451,
)


_AGENTREWARDREPLY = _descriptor.Descriptor(
  name='AgentRewardReply',
  full_name='cogment.AgentRewardReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=453,
  serialized_end=471,
)


_AGENTMESSAGEREQUEST = _descriptor.Descriptor(
  name='AgentMessageRequest',
  full_name='cogment.AgentMessageRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='messages', full_name='cogment.AgentMessageRequest.messages', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=473,
  serialized_end=530,
)


_AGENTMESSAGEREPLY = _descriptor.Descriptor(
  name='AgentMessageReply',
  full_name='cogment.AgentMessageReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=532,
  serialized_end=551,
)


_AGENTENDREQUEST = _descriptor.Descriptor(
  name='AgentEndRequest',
  full_name='cogment.AgentEndRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='final_data', full_name='cogment.AgentEndRequest.final_data', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=553,
  serialized_end=616,
)


_AGENTENDREPLY = _descriptor.Descriptor(
  name='AgentEndReply',
  full_name='cogment.AgentEndReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=618,
  serialized_end=633,
)

_AGENTSTARTREQUEST.fields_by_name['config'].message_type = cogment_dot_api_dot_common__pb2._ACTORCONFIG
_AGENTSTARTREQUEST.fields_by_name['actors_in_trial'].message_type = cogment_dot_api_dot_common__pb2._TRIALACTOR
_AGENTOBSERVATIONREQUEST.fields_by_name['observation'].message_type = cogment_dot_api_dot_common__pb2._OBSERVATION
_AGENTACTIONREPLY.fields_by_name['action'].message_type = cogment_dot_api_dot_common__pb2._ACTION
_AGENTACTIONREPLY.fields_by_name['rewards'].message_type = cogment_dot_api_dot_common__pb2._REWARD
_AGENTACTIONREPLY.fields_by_name['messages'].message_type = cogment_dot_api_dot_common__pb2._MESSAGE
_AGENTREWARDREQUEST.fields_by_name['reward'].message_type = cogment_dot_api_dot_common__pb2._REWARD
_AGENTMESSAGEREQUEST.fields_by_name['messages'].message_type = cogment_dot_api_dot_common__pb2._MESSAGE
_AGENTENDREQUEST.fields_by_name['final_data'].message_type = cogment_dot_api_dot_common__pb2._ACTORPERIODDATA
DESCRIPTOR.message_types_by_name['AgentStartRequest'] = _AGENTSTARTREQUEST
DESCRIPTOR.message_types_by_name['AgentStartReply'] = _AGENTSTARTREPLY
DESCRIPTOR.message_types_by_name['AgentObservationRequest'] = _AGENTOBSERVATIONREQUEST
DESCRIPTOR.message_types_by_name['AgentActionReply'] = _AGENTACTIONREPLY
DESCRIPTOR.message_types_by_name['AgentRewardRequest'] = _AGENTREWARDREQUEST
DESCRIPTOR.message_types_by_name['AgentRewardReply'] = _AGENTREWARDREPLY
DESCRIPTOR.message_types_by_name['AgentMessageRequest'] = _AGENTMESSAGEREQUEST
DESCRIPTOR.message_types_by_name['AgentMessageReply'] = _AGENTMESSAGEREPLY
DESCRIPTOR.message_types_by_name['AgentEndRequest'] = _AGENTENDREQUEST
DESCRIPTOR.message_types_by_name['AgentEndReply'] = _AGENTENDREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AgentStartRequest = _reflection.GeneratedProtocolMessageType('AgentStartRequest', (_message.Message,), {
  'DESCRIPTOR' : _AGENTSTARTREQUEST,
  '__module__' : 'cogment.api.agent_pb2'
  # @@protoc_insertion_point(class_scope:cogment.AgentStartRequest)
  })
_sym_db.RegisterMessage(AgentStartRequest)

AgentStartReply = _reflection.GeneratedProtocolMessageType('AgentStartReply', (_message.Message,), {
  'DESCRIPTOR' : _AGENTSTARTREPLY,
  '__module__' : 'cogment.api.agent_pb2'
  # @@protoc_insertion_point(class_scope:cogment.AgentStartReply)
  })
_sym_db.RegisterMessage(AgentStartReply)

AgentObservationRequest = _reflection.GeneratedProtocolMessageType('AgentObservationRequest', (_message.Message,), {
  'DESCRIPTOR' : _AGENTOBSERVATIONREQUEST,
  '__module__' : 'cogment.api.agent_pb2'
  # @@protoc_insertion_point(class_scope:cogment.AgentObservationRequest)
  })
_sym_db.RegisterMessage(AgentObservationRequest)

AgentActionReply = _reflection.GeneratedProtocolMessageType('AgentActionReply', (_message.Message,), {
  'DESCRIPTOR' : _AGENTACTIONREPLY,
  '__module__' : 'cogment.api.agent_pb2'
  # @@protoc_insertion_point(class_scope:cogment.AgentActionReply)
  })
_sym_db.RegisterMessage(AgentActionReply)

AgentRewardRequest = _reflection.GeneratedProtocolMessageType('AgentRewardRequest', (_message.Message,), {
  'DESCRIPTOR' : _AGENTREWARDREQUEST,
  '__module__' : 'cogment.api.agent_pb2'
  # @@protoc_insertion_point(class_scope:cogment.AgentRewardRequest)
  })
_sym_db.RegisterMessage(AgentRewardRequest)

AgentRewardReply = _reflection.GeneratedProtocolMessageType('AgentRewardReply', (_message.Message,), {
  'DESCRIPTOR' : _AGENTREWARDREPLY,
  '__module__' : 'cogment.api.agent_pb2'
  # @@protoc_insertion_point(class_scope:cogment.AgentRewardReply)
  })
_sym_db.RegisterMessage(AgentRewardReply)

AgentMessageRequest = _reflection.GeneratedProtocolMessageType('AgentMessageRequest', (_message.Message,), {
  'DESCRIPTOR' : _AGENTMESSAGEREQUEST,
  '__module__' : 'cogment.api.agent_pb2'
  # @@protoc_insertion_point(class_scope:cogment.AgentMessageRequest)
  })
_sym_db.RegisterMessage(AgentMessageRequest)

AgentMessageReply = _reflection.GeneratedProtocolMessageType('AgentMessageReply', (_message.Message,), {
  'DESCRIPTOR' : _AGENTMESSAGEREPLY,
  '__module__' : 'cogment.api.agent_pb2'
  # @@protoc_insertion_point(class_scope:cogment.AgentMessageReply)
  })
_sym_db.RegisterMessage(AgentMessageReply)

AgentEndRequest = _reflection.GeneratedProtocolMessageType('AgentEndRequest', (_message.Message,), {
  'DESCRIPTOR' : _AGENTENDREQUEST,
  '__module__' : 'cogment.api.agent_pb2'
  # @@protoc_insertion_point(class_scope:cogment.AgentEndRequest)
  })
_sym_db.RegisterMessage(AgentEndRequest)

AgentEndReply = _reflection.GeneratedProtocolMessageType('AgentEndReply', (_message.Message,), {
  'DESCRIPTOR' : _AGENTENDREPLY,
  '__module__' : 'cogment.api.agent_pb2'
  # @@protoc_insertion_point(class_scope:cogment.AgentEndReply)
  })
_sym_db.RegisterMessage(AgentEndReply)



_AGENTENDPOINT = _descriptor.ServiceDescriptor(
  name='AgentEndpoint',
  full_name='cogment.AgentEndpoint',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=636,
  serialized_end=1066,
  methods=[
  _descriptor.MethodDescriptor(
    name='OnStart',
    full_name='cogment.AgentEndpoint.OnStart',
    index=0,
    containing_service=None,
    input_type=_AGENTSTARTREQUEST,
    output_type=_AGENTSTARTREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='OnObservation',
    full_name='cogment.AgentEndpoint.OnObservation',
    index=1,
    containing_service=None,
    input_type=_AGENTOBSERVATIONREQUEST,
    output_type=_AGENTACTIONREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='OnReward',
    full_name='cogment.AgentEndpoint.OnReward',
    index=2,
    containing_service=None,
    input_type=_AGENTREWARDREQUEST,
    output_type=_AGENTREWARDREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='OnMessage',
    full_name='cogment.AgentEndpoint.OnMessage',
    index=3,
    containing_service=None,
    input_type=_AGENTMESSAGEREQUEST,
    output_type=_AGENTMESSAGEREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='OnEnd',
    full_name='cogment.AgentEndpoint.OnEnd',
    index=4,
    containing_service=None,
    input_type=_AGENTENDREQUEST,
    output_type=_AGENTENDREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Version',
    full_name='cogment.AgentEndpoint.Version',
    index=5,
    containing_service=None,
    input_type=cogment_dot_api_dot_common__pb2._VERSIONREQUEST,
    output_type=cogment_dot_api_dot_common__pb2._VERSIONINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_AGENTENDPOINT)

DESCRIPTOR.services_by_name['AgentEndpoint'] = _AGENTENDPOINT

# @@protoc_insertion_point(module_scope)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from typing import List

import cogment.api.common_pb2 as common_api
from cogment.environment import ENVIRONMENT_ACTOR_NAME
from cogment.errors import Error


class Trial:
    class Environment:
        def __init__(self):
            self.name = ENVIRONMENT_ACTOR_NAME
            self._messages = deque()

        def add_prepared_message(self, payload):
            self._messages.append(payload)

        def get_prepared_messages(self, sender_name):
            while len(self._messages) > 0:
                payload = self._messages.popleft()
                message = common_api.Message(tick_id=-1, sender_name=sender_name, receiver_name=self.name)
                if payload is not None:
                    message.payload.Pack(payload)
                yield message

    class Actor(Environment):
        def __init__(self, name, actor_class):
            super().__init__()
            self.name = name
            self.actor_class = actor_class
            self._reward_data = deque()

        def add_prepared_reward_data(self, tick_id, value, confidence, user_data):
            self._reward_data.append((tick_id, value, confidence, user_data))

        def get_prepared_rewards(self):
            while len(self._reward_data) > 0:
                (tick_id, reward_value, confidence, user_data) = self._reward_data.popleft()
                source = common_api.RewardSource(
                    # sender_name is not needed for sending rewards
                    value=reward_value,
                    confidence=confidence
                )
                if user_data is not None:
                    source.user_data.Pack(user_data)

                reward = common_api.Reward(
                    value=reward_value,  # This will be overwritten by orchestrator before delivery
                    receiver_name=self.name,
                    tick_id=tick_id,
                )
                reward.sources.append(source)
                yield reward

    def __init__(self, id, actors_in_trial, cog_settings):
        self.id = id
        self.over = False
        self.cog_settings = cog_settings
        self.tick_id = -1
        self.environment = self.Environment()

        self._actions = None  # Managed externally
        self._actions_by_actor_id = None  # Managed externally

        self.__actor_by_name = {}

        self.actors = []
        self.actor_counts = []
        self._add_actors(actors_in_trial)
        self._add_actor_counts()

    def _add_actors(self, actors_in_trial):
        for actor_in_trial in actors_in_trial:
            if actor_in_trial.actor_class not in self.cog_settings.actor_classes:
                raise Error(f"class '{actor_in_trial.actor_class}' of actor '{actor_in_trial.name}' can not be found.")
            actor_class = self.cog_settings.actor_classes[actor_in_trial.actor_class]
            actor = self.Actor(
                name=actor_in_trial.name,
                actor_class=actor_class
            )
            self.actors.append(actor)

    def _add_actor_counts(self):
        class_list = self.cog_settings.actor_classes._actor_classes_list
        self.actor_counts = [0] * len(class_list)
        for class_index, class_member in enumerate(class_list):
            for actor in self.actors:
                if class_member.id == actor.actor_class.id:
                    self.actor_counts[class_index] += 1

    def get_actors(self, pattern_list: List[str]):
        if not isinstance(pattern_list, list):
            raise TypeError(f"The pattern_list must be a list: {type(pattern_list)}")
        matched_actors = []
        for actor in self.actors:
            for pattern in pattern_list:
                if not isinstance(pattern, str):
                    raise TypeError(f"The pattern must be a string: {type(pattern)}")
                if pattern == "*":
                    matched_actors.append(actor)
                    break
                else:
                    if "." not in pattern:
                        if actor.name == pattern:
                            matched_actors.append(actor)
                            break
                    else:
                        [class_name, actor_name] = pattern.split(".")
                        if actor_name == actor.name:
                            matched_actors.append(actor)
                            break
                        elif actor_name == "*":
                            if actor.actor_class.id == class_name:
                                matched_actors.append(actor)
                                break

        return matched_actors

    # TODO: Add source actor to set in reward
    def add_reward(self, value, confidence, to, tick_id, user_data):
        for dest_actor in self.get_actors(pattern_list=to):
            dest_actor.add_prepared_reward_data(
                tick_id=tick_id,
                value=value,
                confidence=confidence,
                user_data=user_data
            )

    def _gather_all_rewards(self):
        for actor in self.actors:
            for reward in actor.get_prepared_rewards():
                yield reward

    # We default `to` so users can provide a `to_environment=True` without a `to` parameter
    def send_message(self, user_data, to=[], to_environment=False):
        if to_environment:
            self.environment.add_prepared_message(payload=user_data)
        for actor in self.get_actors(pattern_list=to):
            actor.add_prepared_message(payload=user_data)

    def _gather_all_messages(self, source_name):
        for actor in self.actors:
            for message in actor.get_prepared_messages(source_name):
                yield message

        for message in self.environment.get_prepared_messages(source_name):
            yield message



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from cogment.api import agent_pb2 as cogment_dot_api_dot_agent__pb2
from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2


class AgentEndpointStub(object):
    """This is the connection between a model application and the orchestrator.
    The service is hosted on the model.
    Expected metadata for all calls except `Version`:
    - trial-id: The id of the trial
    - actor-name: The name of the actor within the trial
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.OnStart = channel.unary_unary(
                '/cogment.AgentEndpoint/OnStart',
                request_serializer=cogment_dot_api_dot_agent__pb2.AgentStartRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_agent__pb2.AgentStartReply.FromString,
                )
        self.OnObservation = channel.stream_stream(
                '/cogment.AgentEndpoint/OnObservation',
                request_serializer=cogment_dot_api_dot_agent__pb2.AgentObservationRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_agent__pb2.AgentActionReply.FromString,
                )
        self.OnReward = channel.unary_unary(
                '/cogment.AgentEndpoint/OnReward',
                request_serializer=cogment_dot_api_dot_agent__pb2.AgentRewardRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_agent__pb2.AgentRewardReply.FromString,
                )
        self.OnMessage = channel.unary_unary(
                '/cogment.AgentEndpoint/OnMessage',
                request_serializer=cogment_dot_api_dot_agent__pb2.AgentMessageRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_agent__pb2.AgentMessageReply.FromString,
                )
        self.OnEnd = channel.unary_unary(
                '/cogment.AgentEndpoint/OnEnd',
                request_serializer=cogment_dot_api_dot_agent__pb2.AgentEndRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_agent__pb2.AgentEndReply.FromString,
                )
        self.Version = channel.unary_unary(
                '/cogment.AgentEndpoint/Version',
                request_serializer=cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
                )


class AgentEndpointServicer(object):
    """This is the connection between a model application and the orchestrator.
    The service is hosted on the model.
    Expected metadata for all calls except `Version`:
    - trial-id: The id of the trial
    - actor-name: The name of the actor within the trial
    """

    def OnStart(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OnObservation(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OnReward(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OnMessage(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OnEnd(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Version(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AgentEndpointServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'OnStart': grpc.unary_unary_rpc_method_handler(
                    servicer.OnStart,
                    request_deserializer=cogment_dot_api_dot_agent__pb2.AgentStartRequest.FromString,
                    response_serializer=cogment_dot_api_dot_agent__pb2.AgentStartReply.SerializeToString,
            ),
            'OnObservation': grpc.stream_stream_rpc_method_handler(
                    servicer.OnObservation,
                    request_deserializer=cogment_dot_api_dot_agent__pb2.AgentObservationRequest.FromString,
                    response_serializer=cogment_dot_api_dot_agent__pb2.AgentActionReply.SerializeToString,
            ),
            'OnReward': grpc.unary_unary_rpc_method_handler(
                    servicer.OnReward,
                    request_deserializer=cogment_dot_api_dot_agent__pb2.AgentRewardRequest.FromString,
                    response_serializer=cogment_dot_api_dot_agent__pb2.AgentRewardReply.SerializeToString,
            ),
            'OnMessage': grpc.unary_unary_rpc_method_handler(
                    servicer.OnMessage,
                    request_deserializer=cogment_dot_api_dot_agent__pb2.AgentMessageRequest.FromString,
                    response_serializer=cogment_dot_api_dot_agent__pb2.AgentMessageReply.SerializeToString,
            ),
            'OnEnd': grpc.unary_unary_rpc_method_handler(
                    servicer.OnEnd,
                    request_deserializer=cogment_dot_api_dot_agent__pb2.AgentEndRequest.FromString,
                    response_serializer=cogment_dot_api_dot_agent__pb2.AgentEndReply.SerializeToString,
            ),
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=cogment_dot_api_dot_common__pb2.VersionRequest.FromString,
                    response_serializer=cogment_dot_api_dot_common__pb2.VersionInfo.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cogment.AgentEndpoint', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class AgentEndpoint(object):
    """This is the connection between a model application and the orchestrator.
    The service is hosted on the model.
    Expected metadata for all calls except `Version`:
    - trial-id: The id of the trial
    - actor-name: The name of the actor within the trial
    """

    @staticmethod
    def OnStart(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.AgentEndpoint/OnStart',
            cogment_dot_api_dot_agent__pb2.AgentStartRequest.SerializeToString,
            cogment_dot_api_dot_agent__pb2.AgentStartReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def OnObservation(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/cogment.AgentEndpoint/OnObservation',
            cogment_dot_api_dot_agent__pb2.AgentObservationRequest.SerializeToString,
            cogment_dot_api_dot_agent__pb2.AgentActionReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def OnReward(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.AgentEndpoint/OnReward',
            cogment_dot_api_dot_agent__pb2.AgentRewardRequest.SerializeToString,
            cogment_dot_api_dot_agent__pb2.AgentRewardReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def OnMessage(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.AgentEndpoint/OnMessage',
            cogment_dot_api_dot_agent__pb2.AgentMessageRequest.SerializeToString,
            cogment_dot_api_dot_agent__pb2.AgentMessageReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def OnEnd(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.AgentEndpoint/OnEnd',
            cogment_dot_api_dot_agent__pb2.AgentEndRequest.SerializeToString,
            cogment_dot_api_dot_agent__pb2.AgentEndReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.AgentEndpoint/Version',
            cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
            cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import importlib
import logging
import traceback
from abc import ABC
from cogment.session import Session

import cogment.api.orchestrator_pb2 as orchestrator_api
import cogment.api.common_pb2 as common_api


_OBSERVATION = "observation"
_FINAL_DATA = "final_data"
_MESSAGE = "message"
_REWARD = "reward"


class ActorClass:
    def __init__(
        self,
        id,
        config_type,
        action_space,
        observation_space,
        observation_delta,
        observation_delta_apply_fn,
        feedback_space,
    ):
        self.id = id
        self.config_type = config_type
        self.action_space = action_space
        self.observation_space = observation_space
        self.observation_delta = observation_delta
        self.observation_delta_apply_fn = observation_delta_apply_fn
        self.feedback_space = feedback_space


class ActorClassList:
    def __init__(self, *args):
        self._actor_classes_list = list(args)

        for a_c in args:
            setattr(self, a_c.id, a_c)

    def __iter__(self):
        return iter(self._actor_classes_list)

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        return getattr(self, key)

    def get_class_by_index(self, index):
        return self._actor_classes_list[index]


class RecvReward:
    def __init__(self):
        self.tick_id = -1
        self.value = 0
        self._sources = None

    def _set_all(self, reward):
        self.tick_id = reward.tick_id
        self.value = reward.value
        self._sources = reward.sources

    def get_nb_sources(self):
        return len(self._sources)

    def all_sources(self):
        assert self._sources
        for src in self._sources:
            yield (src.value, src.confidence, src.sender_name, src.user_data)

    def __str__(self):
        res = f"tick_id = {self.tick_id}, value = {self.value}"
        res += f", sources = {self._sources}"
        return res


class ActorSession(Session):
    """This represents an actor being performed locally."""

    def __init__(self, impl, actor_class, trial, name, impl_name, config):
        super().__init__(trial)
        self.class_name = actor_class.id
        self.name = name
        self.impl_name = impl_name
        self.config = config

        self._actor_class = actor_class
        self._ended = False
        self._action_queue = asyncio.Queue()
        self._task = None  # Task used to call _run()
        self._latest_observation = None
        self._last_event_received = False

        self.__impl = impl
        self.__started = False
        self.__event_queue = None
        self.__obs_future = None

    def start(self):
        assert not self.__started
        assert not self._ended

        self.__event_queue = asyncio.Queue()
        self.__started = True

    async def event_loop(self):
        assert self.__started
        assert not self._ended

        loop_active = not self._last_event_received
        while loop_active:
            try:
                event = await self.__event_queue.get()

            except asyncio.CancelledError as exc:
                logging.debug(f"Coroutine for actor [{self.name}] cancelled while waiting for an event")
                break

            self._last_event_received = _FINAL_DATA in event
            keep_looping = yield event
            self.__event_queue.task_done()
            loop_active = (keep_looping is None or bool(keep_looping)) and not self._last_event_received
            if not loop_active:
                if self._last_event_received:
                    logging.debug(f"Last event received, exiting event loop for actor [{self.name}]")
                else:
                    logging.debug(f"End of event loop for actor [{self.name}] requested by user")

        logging.debug(f"Exiting actor [{self.name}] event loop generator")

    def do_action(self, action):
        assert self.__started
        self._action_queue.put_nowait(action)

    async def _retrieve_action(self):
        action = await self._action_queue.get()
        self._action_queue.task_done()
        return action

    async def _end(self, package):
        logging.debug(f"Actor [{self.name}] received final data")
        if not self._ended:
            self._ended = True

        if self.__event_queue:
            std_messages = []
            for msg in package.messages:
                std_messages.append((msg.sender_name(), msg.payload()))
            package.messages = std_messages

            event = {_FINAL_DATA : package}
            await self.__event_queue.put(event)
        else:
            logging.warning(f"Actor [{self.name}] received final data that it was unable to handle.")

    def _new_observation(self, obs):
        if not self.__started or self._ended:
            return

        if self.__event_queue:
            event = {_OBSERVATION : obs}
            self.__event_queue.put_nowait(event)
        else:
            logging.warning(f"Actor [{self.name}] received an observation that it was unable to handle.")

    def _new_reward(self, reward):
        logging.debug(f"Actor [{self.name}] received a reward")
        if not self.__started or self._ended:
            return

        if self.__event_queue:
            event = {_REWARD : reward}
            self.__event_queue.put_nowait(event)
        else:
            logging.warning(f"Actor [{self.name}] received a reward that it was unable to handle.")

    def _new_message(self, message):
        logging.debug(f"Actor [{self.name}] received a message")
        if not self.__started or self._ended:
            return

        if self.__event_queue:
            event = {_MESSAGE : (message.sender_name, message.payload)}
            self.__event_queue.put_nowait(event)
        else:
            logging.warning(f"Actor [{self.name}] received a message that it was unable to handle.")

    async def _run(self):
        try:
            await self.__impl(self)

        except asyncio.CancelledError:
            logging.debug(f"Actor [{self.name}] implementation coroutine cancelled")

        except Exception:
            logging.error(f"An exception occured in user actor implementation [{self.impl_name}]:\n"
                          f"{traceback.format_exc()}")
            raise


class _ServedActorSession(ActorSession):
    def __init__(self, impl, actor_class, trial, name, impl_name, config):
        super().__init__(impl, actor_class, trial, name, impl_name, config)

    def add_reward(self, value, confidence, to, tick_id=-1, user_data=None):
        assert self._trial is not None
        self._trial.add_reward(value, confidence, to, tick_id, user_data)

    def send_message(self, user_data, to, to_environment=False):
        assert self._trial is not None
        self._trial.send_message(user_data, to, to_environment)


class _ClientActorSession(ActorSession):
    def __init__(self, impl, actor_class, trial, name, impl_name, config, actor_stub):
        super().__init__(impl, actor_class, trial, name, impl_name, config)
        self._actor_sub = actor_stub

    def add_reward(self, value, confidence, to, tick_id=-1, user_data=None):
        request = orchestrator_api.TrialRewardRequest()
        for dest_actor in self._trial.get_actors(pattern_list=to):
            reward = common_api.Reward(receiver_name=dest_actor.name, tick_id=-1, value=value)
            reward_source = common_api.RewardSource(sender_name=self.name, value=value, confidence=confidence)
            if user_data is not None:
                reward_source.user_data.Pack(user_data)
            reward.sources.append(reward_source)
            request.rewards.append(reward)
            metadata = (("trial-id", self.get_trial_id()), ("actor-name", self.name))
            self._actor_sub.SendReward(request=request, metadata=metadata)

    def send_message(self, user_data, to, to_environment=False):
        message_req = orchestrator_api.TrialMessageRequest()
        for dest_actor in self._trial.get_actors(pattern_list=to):
            message = common_api.Message(tick_id=-1, receiver_name=dest_actor.name)
            if user_data is not None:
                message.payload.Pack(user_data)
            message_req.messages.append(message)
            metadata = (("trial-id", self.get_trial_id()), ("actor-name", self.name))
            self._actor_sub.SendMessage(request=message_req, metadata=metadata)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod


class Session(ABC):
    class ActiveActor:
        def __init__(self, actor_name, actor_class_name):
            self.actor_name = actor_name
            self.actor_class_name = actor_class_name

    def __init__(self, trial):
        self._trial = trial

        # Pre-compute since it will be used regularly
        self._active_actors = [self.ActiveActor(actor_name=actor.name, actor_class_name=actor.actor_class.id)
                             for actor in trial.actors]

    def get_trial_id(self):
        assert self._trial is not None
        return self._trial.id

    def get_tick_id(self):
        assert self._trial is not None
        return self._trial.tick_id

    def is_trial_over(self):
        assert self._trial is not None
        return self._trial.over

    def get_active_actors(self):
        assert self._trial is not None
        return self._active_actors

    @abstractmethod
    def add_reward(self, value, confidence, to, tick_id=-1, user_data=None):
        pass

    @abstractmethod
    def send_message(self, user_data, to, to_environment=False):
        pass



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cogment.api.common_pb2 as common_api
from cogment.version import __version__
from cogment.delta_encoding import DecodeObservationData

import logging
import importlib
import grpc


def list_versions():
    reply = common_api.VersionInfo()
    reply.versions.add(name='cogment_sdk', version=__version__)
    reply.versions.add(name='grpc', version=grpc.__version__)

    return reply


def raw_params_to_user_params(params, settings):
    trial_config = None
    if params.HasField("trial_config"):
        trial_config = settings.trial.config_type()
        trial_config.ParseFromString(params.trial_config.content)

    env_config = None
    if(params.environment.HasField("config")):
        env_config = settings.environment.config_type()
        env_config.ParseFromString(params.environment.config.content)

    environment = {
        "endpoint": params.environment.endpoint,
        "config": env_config
    }

    actors = []
    for actor in params.actors:
        actor_config = None

        if actor.HasField("config"):
            a_c = settings.actor_classes.__getattribute__(actor.actor_class)
            actor_config = a_c.config_type()
            actor_config.ParseFromString(actor.config.content)

        actor_data = {
            "name": actor.name,
            "actor_class": actor.actor_class,
            "endpoint": actor.endpoint,
            "implementation": actor.implementation,
            "config": actor_config
        }
        actors.append(actor_data)

    return {
        "trial_config": trial_config,
        "environment": environment,
        "actors": actors,
        "max_steps": params.max_steps,
        "max_inactivity": params.max_inactivity
    }


def user_params_to_raw_params(params, settings):
    result = common_api.TrialParams()

    result.max_steps = params["max_steps"]
    result.max_inactivity = params["max_inactivity"]

    if params["trial_config"] is not None:
        result.trial_config.content = params["trial_config"].SerializeToString()

    result.environment.endpoint = params["environment"]["endpoint"]
    if params["environment"]["config"] is not None:
        result.environment.config.content = params["environment"]["config"].SerializeToString()

    for actor_data in params["actors"]:
        actor_pb = result.actors.add()
        actor_pb.name = actor_data["name"]
        actor_pb.actor_class = actor_data["actor_class"]
        actor_pb.endpoint = actor_data["endpoint"]
        actor_pb.implementation = actor_data["implementation"]
        if actor_data["config"] is not None:
            actor_pb.config.content = actor_data["config"].SerializeToString()

    return result


class DecodeData():

    def __init__(self, trial_params, cog_settings):
        self.__cog_settings = cog_settings
        self.last_obs = []

        actor_classes_list = [
            actor.id for actor in self.__cog_settings.actor_classes]
        trial_actor_list = [actor.actor_class for actor in trial_params.actors]
        self.actor_counts = [0] * len(actor_classes_list)
        for index, actor_class in enumerate(actor_classes_list):
            self.actor_counts[index] += trial_actor_list.count(actor_class)

        for ac_index, actor_class in enumerate(self.__cog_settings.actor_classes):
            count = self.actor_counts[ac_index]
            self.last_obs.extend([None] * count)

    def decode_datasample(self, sample):

        actor_index = 0
        for ac_index, actor_class in enumerate(self.__cog_settings.actor_classes):
            count = self.actor_counts[ac_index]
            for _ in range(count):
                obs_id = sample.observations.actors_map[actor_index]
                obs_data = sample.observations.observations[obs_id]

                obs = DecodeObservationData(
                    actor_class, obs_data, self.last_obs[actor_index])
                self.last_obs[actor_index] = obs

                actor_index += 1

        action_list = []
        for act_data in sample.actions:
            action = self.__cog_settings.data_pb.Action()
            action.ParseFromString(act_data.content)
            action_list.append(action)

        reward_list = []
        for rwd in sample.rewards:
            reward_list.append((rwd.value, rwd.confidence))

        message_list = []
        for messages in sample.messages:
            sub_msg_list = []
            for message in messages.messages:

                class_type = message.payload.type_url.split('.')
                user_data = getattr(importlib.import_module(
                    self.__cog_settings.protolib), class_type[-1])()
                message.payload.Unpack(user_data)

                sub_msg_list.append((message.sender_name, user_data))

            message_list.append(sub_msg_list)

        return self.last_obs, action_list, reward_list, message_list



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import traceback
import grpc
import grpc.experimental.aio

import cogment.api.orchestrator_pb2 as orchestrator_api
import cogment.api.orchestrator_pb2_grpc as grpc_api
from cogment.trial import Trial
from cogment.control import _ServedControlSession


class ControlServicer:
    def __init__(self, cog_settings, endpoint):
        self.cog_settings = cog_settings
        if endpoint.private_key is None:
            channel = grpc.experimental.aio.insecure_channel(endpoint.url)
        else:
            if endpoint.root_certificates:
                root = bytes(endpoint.root_certificates, "utf-8")
            else:
                root = None
            if endpoint.private_key:
                key = bytes(endpoint.private_key, "utf-8")
            else:
                key = None
            if endpoint.certificate_chain:
                certs = bytes(endpoint.certificate_chain, "utf-8")
            else:
                certs = None
            creds = grpc.ssl_channel_credentials(root, key, certs)
            channel = grpc.experimental.aio.secure_channel(endpoint.url, creds)

        self.lifecycle_stub = grpc_api.TrialLifecycleStub(channel)

    async def run(self, user_id, impl, trial_config):
        req = orchestrator_api.TrialStartRequest()
        req.user_id = user_id
        if trial_config is not None:
            req.config.content = trial_config.SerializeToString()

        rep = await self.lifecycle_stub.StartTrial(req)
        trial = Trial(rep.trial_id, rep.actors_in_trial, self.cog_settings)

        try:
            await impl(_ServedControlSession(trial, self.lifecycle_stub))
            logging.debug(f"Control implementation returned")

        except asyncio.CancelledError:
            logging.debug("Control implementation coroutine cancelled")

        except Exception:
            logging.error(f"An exception occured in user control implementation:\n{traceback.format_exc()}")
            raise



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2
from cogment.api import orchestrator_pb2 as cogment_dot_api_dot_orchestrator__pb2


class TrialLifecycleStub(object):
    """This service is used to manage Trial lifecycle.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.StartTrial = channel.unary_unary(
                '/cogment.TrialLifecycle/StartTrial',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialStartRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialStartReply.FromString,
                )
        self.TerminateTrial = channel.unary_unary(
                '/cogment.TrialLifecycle/TerminateTrial',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TerminateTrialRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TerminateTrialReply.FromString,
                )
        self.GetTrialInfo = channel.unary_unary(
                '/cogment.TrialLifecycle/GetTrialInfo',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialInfoRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialInfoReply.FromString,
                )
        self.WatchTrials = channel.unary_stream(
                '/cogment.TrialLifecycle/WatchTrials',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialListRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialListEntry.FromString,
                )
        self.Version = channel.unary_unary(
                '/cogment.TrialLifecycle/Version',
                request_serializer=cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
                )


class TrialLifecycleServicer(object):
    """This service is used to manage Trial lifecycle.
    """

    def StartTrial(self, request, context):
        """Begin a new trial.
        Expected headers: None
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TerminateTrial(self, request, context):
        """Terminate an existing trial.
        Expected headers: 
        - trial-id
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTrialInfo(self, request, context):
        """Get extra information about a specific trial.
        Expected headers: 
        - trial-id <optional>
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def WatchTrials(self, request, context):
        """Get information about the trials currently running in the orchestrator.
        The stream will be prefixed by all the currently filtered existing trials,
        and will return a new message whenever a trial enters a filtered state.
        Expected headers: None
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Version(self, request, context):
        """Expected headers: None
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TrialLifecycleServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'StartTrial': grpc.unary_unary_rpc_method_handler(
                    servicer.StartTrial,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialStartRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialStartReply.SerializeToString,
            ),
            'TerminateTrial': grpc.unary_unary_rpc_method_handler(
                    servicer.TerminateTrial,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TerminateTrialRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TerminateTrialReply.SerializeToString,
            ),
            'GetTrialInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTrialInfo,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialInfoRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialInfoReply.SerializeToString,
            ),
            'WatchTrials': grpc.unary_stream_rpc_method_handler(
                    servicer.WatchTrials,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialListRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialListEntry.SerializeToString,
            ),
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=cogment_dot_api_dot_common__pb2.VersionRequest.FromString,
                    response_serializer=cogment_dot_api_dot_common__pb2.VersionInfo.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cogment.TrialLifecycle', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class TrialLifecycle(object):
    """This service is used to manage Trial lifecycle.
    """

    @staticmethod
    def StartTrial(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.TrialLifecycle/StartTrial',
            cogment_dot_api_dot_orchestrator__pb2.TrialStartRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TrialStartReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def TerminateTrial(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.TrialLifecycle/TerminateTrial',
            cogment_dot_api_dot_orchestrator__pb2.TerminateTrialRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TerminateTrialReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTrialInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.TrialLifecycle/GetTrialInfo',
            cogment_dot_api_dot_orchestrator__pb2.TrialInfoRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TrialInfoReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def WatchTrials(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/cogment.TrialLifecycle/WatchTrials',
            cogment_dot_api_dot_orchestrator__pb2.TrialListRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TrialListEntry.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.TrialLifecycle/Version',
            cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
            cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class ClientActorStub(object):
    """This service is used by an actor within a trial, using the "client" API.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.JoinTrial = channel.unary_unary(
                '/cogment.ClientActor/JoinTrial',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialJoinRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialJoinReply.FromString,
                )
        self.ActionStream = channel.stream_stream(
                '/cogment.ClientActor/ActionStream',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialActionRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialActionReply.FromString,
                )
        self.Heartbeat = channel.unary_unary(
                '/cogment.ClientActor/Heartbeat',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialHeartbeatRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialHeartbeatReply.FromString,
                )
        self.SendReward = channel.unary_unary(
                '/cogment.ClientActor/SendReward',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialRewardRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialRewardReply.FromString,
                )
        self.SendMessage = channel.unary_unary(
                '/cogment.ClientActor/SendMessage',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialMessageRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialMessageReply.FromString,
                )
        self.Version = channel.unary_unary(
                '/cogment.ClientActor/Version',
                request_serializer=cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
                )


class ClientActorServicer(object):
    """This service is used by an actor within a trial, using the "client" API.
    """

    def JoinTrial(self, request, context):
        """Expected headers: None
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ActionStream(self, request_iterator, context):
        """Expected headers: 
        - trial-id
        - actor-name
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Heartbeat(self, request, context):
        """Expected headers: 
        - trial-id
        - actor-name
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendReward(self, request, context):
        """Expected headers: 
        - trial-id
        - actor-name
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendMessage(self, request, context):
        """Expected headers: 
        - trial-id
        - actor-name
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Version(self, request, context):
        """Expected headers: None
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ClientActorServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'JoinTrial': grpc.unary_unary_rpc_method_handler(
                    servicer.JoinTrial,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialJoinRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialJoinReply.SerializeToString,
            ),
            'ActionStream': grpc.stream_stream_rpc_method_handler(
                    servicer.ActionStream,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialActionRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialActionReply.SerializeToString,
            ),
            'Heartbeat': grpc.unary_unary_rpc_method_handler(
                    servicer.Heartbeat,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialHeartbeatRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialHeartbeatReply.SerializeToString,
            ),
            'SendReward': grpc.unary_unary_rpc_method_handler(
                    servicer.SendReward,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialRewardRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialRewardReply.SerializeToString,
            ),
            'SendMessage': grpc.unary_unary_rpc_method_handler(
                    servicer.SendMessage,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialMessageRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialMessageReply.SerializeToString,
            ),
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=cogment_dot_api_dot_common__pb2.VersionRequest.FromString,
                    response_serializer=cogment_dot_api_dot_common__pb2.VersionInfo.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cogment.ClientActor', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ClientActor(object):
    """This service is used by an actor within a trial, using the "client" API.
    """

    @staticmethod
    def JoinTrial(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.ClientActor/JoinTrial',
            cogment_dot_api_dot_orchestrator__pb2.TrialJoinRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TrialJoinReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ActionStream(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/cogment.ClientActor/ActionStream',
            cogment_dot_api_dot_orchestrator__pb2.TrialActionRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TrialActionReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Heartbeat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.ClientActor/Heartbeat',
            cogment_dot_api_dot_orchestrator__pb2.TrialHeartbeatRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TrialHeartbeatReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendReward(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.ClientActor/SendReward',
            cogment_dot_api_dot_orchestrator__pb2.TrialRewardRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TrialRewardReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendMessage(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.ClientActor/SendMessage',
            cogment_dot_api_dot_orchestrator__pb2.TrialMessageRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TrialMessageReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.ClientActor/Version',
            cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
            cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cogment.api.orchestrator_pb2 as orchestrator_api

from abc import ABC, abstractmethod
from types import SimpleNamespace


# Future functionality (as a non-participant):
#   - Accept/refuse actor connections
#   - Diconnect actors
#   - Request tick updates?
#   - Request any observation?
#   - Send messages?
#   - Request to receive every message?
class ControlSession(ABC):
    def __init__(self, trial, stub):
        self._trial = trial
        self._lifecycle_stub = stub

    def get_trial_id(self):
        return self._trial.id

    def get_actors(self):
        return [SimpleNamespace(actor_name=actor.name, actor_class_name=actor.actor_class)
                for actor in self._trial.actors]

    async def terminate_trial(self):
        req = orchestrator_api.TerminateTrialRequest()
        metadata = [("trial-id", self._trial.id)]
        await self._lifecycle_stub.TerminateTrial(
            request=req,
            metadata=metadata
        )


class _ServedControlSession(ControlSession):
    def __init__(self, trial, stub):
        super().__init__(trial, stub)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/environment.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='cogment/api/environment.proto',
  package='cogment',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1d\x63ogment/api/environment.proto\x12\x07\x63ogment\x1a\x18\x63ogment/api/common.proto\"x\n\x0eObservationSet\x12\x0f\n\x07tick_id\x18\x01 \x01(\x04\x12\x11\n\ttimestamp\x18\x02 \x01(\x06\x12.\n\x0cobservations\x18\x03 \x03(\x0b\x32\x18.cogment.ObservationData\x12\x12\n\nactors_map\x18\x04 \x03(\x05\"\x1c\n\tActionSet\x12\x0f\n\x07\x61\x63tions\x18\x01 \x03(\x0c\"~\n\x0f\x45nvStartRequest\x12\x11\n\timpl_name\x18\x01 \x01(\t\x12*\n\x06\x63onfig\x18\x02 \x01(\x0b\x32\x1a.cogment.EnvironmentConfig\x12,\n\x0f\x61\x63tors_in_trial\x18\x03 \x03(\x0b\x32\x13.cogment.TrialActor\"A\n\rEnvStartReply\x12\x30\n\x0fobservation_set\x18\x01 \x01(\x0b\x32\x17.cogment.ObservationSet\"W\n\x10\x45nvActionRequest\x12&\n\naction_set\x18\x01 \x01(\x0b\x32\x12.cogment.ActionSet\x12\x1b\n\x13reply_with_snapshot\x18\x02 \x01(\x08\"\x9e\x01\n\x0e\x45nvActionReply\x12\x30\n\x0fobservation_set\x18\x01 \x01(\x0b\x32\x17.cogment.ObservationSet\x12 \n\x07rewards\x18\x02 \x03(\x0b\x32\x0f.cogment.Reward\x12\"\n\x08messages\x18\x03 \x03(\x0b\x32\x10.cogment.Message\x12\x14\n\x0c\x66inal_update\x18\x04 \x01(\x08\"7\n\x11\x45nvMessageRequest\x12\"\n\x08messages\x18\x01 \x03(\x0b\x32\x10.cogment.Message\"\x11\n\x0f\x45nvMessageReply2\xda\x02\n\x13\x45nvironmentEndpoint\x12=\n\x07OnStart\x12\x18.cogment.EnvStartRequest\x1a\x16.cogment.EnvStartReply\"\x00\x12\x44\n\x08OnAction\x12\x19.cogment.EnvActionRequest\x1a\x17.cogment.EnvActionReply\"\x00(\x01\x30\x01\x12\x43\n\tOnMessage\x12\x1a.cogment.EnvMessageRequest\x1a\x18.cogment.EnvMessageReply\"\x00\x12=\n\x05OnEnd\x12\x19.cogment.EnvActionRequest\x1a\x17.cogment.EnvActionReply\"\x00\x12:\n\x07Version\x12\x17.cogment.VersionRequest\x1a\x14.cogment.VersionInfo\"\x00\x62\x06proto3'
  ,
  dependencies=[cogment_dot_api_dot_common__pb2.DESCRIPTOR,])




_OBSERVATIONSET = _descriptor.Descriptor(
  name='ObservationSet',
  full_name='cogment.ObservationSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tick_id', full_name='cogment.ObservationSet.tick_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='cogment.ObservationSet.timestamp', index=1,
      number=2, type=6, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='observations', full_name='cogment.ObservationSet.observations', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actors_map', full_name='cogment.ObservationSet.actors_map', index=3,
      number=4, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=68,
  serialized_end=188,
)


_ACTIONSET = _descriptor.Descriptor(
  name='ActionSet',
  full_name='cogment.ActionSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='actions', full_name='cogment.ActionSet.actions', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=190,
  serialized_end=218,
)


_ENVSTARTREQUEST = _descriptor.Descriptor(
  name='EnvStartRequest',
  full_name='cogment.EnvStartRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='impl_name', full_name='cogment.EnvStartRequest.impl_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='config', full_name='cogment.EnvStartRequest.config', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actors_in_trial', full_name='cogment.EnvStartRequest.actors_in_trial', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=220,
  serialized_end=346,
)


_ENVSTARTREPLY = _descriptor.Descriptor(
  name='EnvStartReply',
  full_name='cogment.EnvStartReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='observation_set', full_name='cogment.EnvStartReply.observation_set', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=348,
  serialized_end=413,
)


_ENVACTIONREQUEST = _descriptor.Descriptor(
  name='EnvActionRequest',
  full_name='cogment.EnvActionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='action_set', full_name='cogment.EnvActionRequest.action_set', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reply_with_snapshot', full_name='cogment.EnvActionRequest.reply_with_snapshot', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=415,
  serialized_end=502,
)


_ENVACTIONREPLY = _descriptor.Descriptor(
  name='EnvActionReply',
  full_name='cogment.EnvActionReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='observation_set', full_name='cogment.EnvActionReply.observation_set', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rewards', full_name='cogment.EnvActionReply.rewards', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='messages', full_name='cogment.EnvActionReply.messages', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='final_update', full_name='cogment.EnvActionReply.final_update', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=505,
  serialized_end=663,
)


_ENVMESSAGEREQUEST = _descriptor.Descriptor(
  name='EnvMessageRequest',
  full_name='cogment.EnvMessageRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='messages', full_name='cogment.EnvMessageRequest.messages', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=665,
  serialized_end=720,
)


_ENVMESSAGEREPLY = _descriptor.Descriptor(
  name='EnvMessageReply',
  full_name='cogment.EnvMessageReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=722,
  serialized_end=739,
)

_OBSERVATIONSET.fields_by_name['observations'].message_type = cogment_dot_api_dot_common__pb2._OBSERVATIONDATA
_ENVSTARTREQUEST.fields_by_name['config'].message_type = cogment_dot_api_dot_common__pb2._ENVIRONMENTCONFIG
_ENVSTARTREQUEST.fields_by_name['actors_in_trial'].message_type = cogment_dot_api_dot_common__pb2._TRIALACTOR
_ENVSTARTREPLY.fields_by_name['observation_set'].message_type = _OBSERVATIONSET
_ENVACTIONREQUEST.fields_by_name['action_set'].message_type = _ACTIONSET
_ENVACTIONREPLY.fields_by_name['observation_set'].message_type = _OBSERVATIONSET
_ENVACTIONREPLY.fields_by_name['rewards'].message_type = cogment_dot_api_dot_common__pb2._REWARD
_ENVACTIONREPLY.fields_by_name['messages'].message_type = cogment_dot_api_dot_common__pb2._MESSAGE
_ENVMESSAGEREQUEST.fields_by_name['messages'].message_type = cogment_dot_api_dot_common__pb2._MESSAGE
DESCRIPTOR.message_types_by_name['ObservationSet'] = _OBSERVATIONSET
DESCRIPTOR.message_types_by_name['ActionSet'] = _ACTIONSET
DESCRIPTOR.message_types_by_name['EnvStartRequest'] = _ENVSTARTREQUEST
DESCRIPTOR.message_types_by_name['EnvStartReply'] = _ENVSTARTREPLY
DESCRIPTOR.message_types_by_name['EnvActionRequest'] = _ENVACTIONREQUEST
DESCRIPTOR.message_types_by_name['EnvActionReply'] = _ENVACTIONREPLY
DESCRIPTOR.message_types_by_name['EnvMessageRequest'] = _ENVMESSAGEREQUEST
DESCRIPTOR.message_types_by_name['EnvMessageReply'] = _ENVMESSAGEREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ObservationSet = _reflection.GeneratedProtocolMessageType('ObservationSet', (_message.Message,), {
  'DESCRIPTOR' : _OBSERVATIONSET,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogment.ObservationSet)
  })
_sym_db.RegisterMessage(ObservationSet)

ActionSet = _reflection.GeneratedProtocolMessageType('ActionSet', (_message.Message,), {
  'DESCRIPTOR' : _ACTIONSET,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogment.ActionSet)
  })
_sym_db.RegisterMessage(ActionSet)

EnvStartRequest = _reflection.GeneratedProtocolMessageType('EnvStartRequest', (_message.Message,), {
  'DESCRIPTOR' : _ENVSTARTREQUEST,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogment.EnvStartRequest)
  })
_sym_db.RegisterMessage(EnvStartRequest)

EnvStartReply = _reflection.GeneratedProtocolMessageType('EnvStartReply', (_message.Message,), {
  'DESCRIPTOR' : _ENVSTARTREPLY,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogment.EnvStartReply)
  })
_sym_db.RegisterMessage(EnvStartReply)

EnvActionRequest = _reflection.GeneratedProtocolMessageType('EnvActionRequest', (_message.Message,), {
  'DESCRIPTOR' : _ENVACTIONREQUEST,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogment.EnvActionRequest)
  })
_sym_db.RegisterMessage(EnvActionRequest)

EnvActionReply = _reflection.GeneratedProtocolMessageType('EnvActionReply', (_message.Message,), {
  'DESCRIPTOR' : _ENVACTIONREPLY,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogment.EnvActionReply)
  })
_sym_db.RegisterMessage(EnvActionReply)

EnvMessageRequest = _reflection.GeneratedProtocolMessageType('EnvMessageRequest', (_message.Message,), {
  'DESCRIPTOR' : _ENVMESSAGEREQUEST,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogment.EnvMessageRequest)
  })
_sym_db.RegisterMessage(EnvMessageRequest)

EnvMessageReply = _reflection.GeneratedProtocolMessageType('EnvMessageReply', (_message.Message,), {
  'DESCRIPTOR' : _ENVMESSAGEREPLY,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogment.EnvMessageReply)
  })
_sym_db.RegisterMessage(EnvMessageReply)



_ENVIRONMENTENDPOINT = _descriptor.ServiceDescriptor(
  name='EnvironmentEndpoint',
  full_name='cogment.EnvironmentEndpoint',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=742,
  serialized_end=1088,
  methods=[
  _descriptor.MethodDescriptor(
    name='OnStart',
    full_name='cogment.EnvironmentEndpoint.OnStart',
    index=0,
    containing_service=None,
    input_type=_ENVSTARTREQUEST,
    output_type=_ENVSTARTREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='OnAction',
    full_name='cogment.EnvironmentEndpoint.OnAction',
    index=1,
    containing_service=None,
    input_type=_ENVACTIONREQUEST,
    output_type=_ENVACTIONREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='OnMessage',
    full_name='cogment.EnvironmentEndpoint.OnMessage',
    index=2,
    containing_service=None,
    input_type=_ENVMESSAGEREQUEST,
    output_type=_ENVMESSAGEREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='OnEnd',
    full_name='cogment.EnvironmentEndpoint.OnEnd',
    index=3,
    containing_service=None,
    input_type=_ENVACTIONREQUEST,
    output_type=_ENVACTIONREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Version',
    full_name='cogment.EnvironmentEndpoint.Version',
    index=4,
    containing_service=None,
    input_type=cogment_dot_api_dot_common__pb2._VERSIONREQUEST,
    output_type=cogment_dot_api_dot_common__pb2._VERSIONINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_ENVIRONMENTENDPOINT)

DESCRIPTOR.services_by_name['EnvironmentEndpoint'] = _ENVIRONMENTENDPOINT

# @@protoc_insertion_point(module_scope)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import grpc.experimental.aio

import cogment.api.orchestrator_pb2 as orchestrator_api
from cogment.actor import _ClientActorSession, RecvReward
import cogment.api.orchestrator_pb2_grpc as grpc_api
from cogment.delta_encoding import DecodeObservationData
from cogment.errors import InvalidRequestError
from cogment.trial import Trial
from types import SimpleNamespace

import asyncio
import logging
import traceback


async def read_observations(client_session, reply_itor):
    try:
        async for reply in reply_itor:
            if not reply.final_data:
                for obs_reply in reply.data.observations:
                    tick_id = obs_reply.tick_id
                    client_session._trial.tick_id = tick_id

                    obs = DecodeObservationData(
                        client_session._actor_class,
                        obs_reply.data,
                        client_session._latest_observation)
                    client_session._latest_observation = obs
                    client_session._new_observation(obs)

                for rew_request in reply.data.rewards:
                    reward = RecvReward()
                    reward._set_all(rew_request)
                    client_session._new_reward(reward)

                for message in reply.data.messages:
                    client_session._new_message(message)

            else:
                client_session._trial.over = True

                package = SimpleNamespace(observations=[], rewards=[], messages=[])
                tick_id = None

                for obs_reply in reply.data.observations:
                    tick_id = obs_reply.tick_id
                    obs = DecodeObservationData(
                        client_session._actor_class,
                        obs_reply.data,
                        client_session._latest_observation)
                    client_session._latest_observation = obs
                    package.observations.append(obs)

                for rew_request in reply.data.rewards:
                    reward = RecvReward()
                    reward._set_all(rew_request)
                    package.rewards.append(reward)

                for msg_request in reply.data.messages:
                    package.messages.append(msg_request)

                if tick_id is not None:
                    client_session._trial.tick_id = tick_id

                await client_session._end(package)

                break

    except asyncio.CancelledError:
        logging.debug(f"Client [{client_session.name}] 'read_observations' coroutine cancelled")

    except grpc.experimental.aio._call.AioRpcError as exc:
        if exc.code() == grpc.StatusCode.UNAVAILABLE:
            logging.error(f"Orchestrator communication lost: [{exc.details()}]")
            logging.debug(f"gRPC Error details: [{exc.debug_error_string()}]")
            client_session._task.cancel()
        else:
            raise

    except Exception:
        logging.error(f"{traceback.format_exc()}")
        raise


async def write_actions(client_session):
    try:
        while True:
            act = await client_session._retrieve_action()

            action_req = orchestrator_api.TrialActionRequest()
            if act is not None:
                action_req.action.content = act.SerializeToString()

            yield action_req

    except asyncio.CancelledError:
        logging.debug(f"Client [{client_session.name}] 'write_actions' coroutine cancelled")

    except Exception:
        logging.error(f"{traceback.format_exc()}")
        raise


class ClientServicer:
    def __init__(self, cog_settings, endpoint):
        self.cog_settings = cog_settings

        if endpoint.private_key is None:
            channel = grpc.experimental.aio.insecure_channel(endpoint.url)
        else:
            if endpoint.root_certificates:
                root = bytes(endpoint.root_certificates, "utf-8")
            else:
                root = None
            if endpoint.private_key:
                key = bytes(endpoint.private_key, "utf-8")
            else:
                key = None
            if endpoint.certificate_chain:
                certs = bytes(endpoint.certificate_chain, "utf-8")
            else:
                certs = None
            creds = grpc.ssl_channel_credentials(root, key, certs)
            channel = grpc.experimental.aio.secure_channel(endpoint.url, creds)

        self._actor_stub = grpc_api.ClientActorStub(channel)

    async def run(self, trial_id, impl, impl_name, actor_classes, actor_name):

        # TODO: Handle properly the multiple actor classes.  Including "all" classes
        #       when the list is empty
        if actor_name is None:
            requested_actor_class = actor_classes[0]
            req = orchestrator_api.TrialJoinRequest(trial_id=trial_id, actor_class=requested_actor_class)
        else:
            req = orchestrator_api.TrialJoinRequest(trial_id=trial_id, actor_name=actor_name)

        reply = await self._actor_stub.JoinTrial(req)

        trial = Trial(reply.trial_id, reply.actors_in_trial, self.cog_settings)

        self_info = None
        for info in reply.actors_in_trial:
            if info.name == reply.actor_name:
                self_info = info
                break
        if self_info is None:
            raise InvalidRequestError(
                f"Unknown actor name: [{reply.actor_name}] not found in [{reply.actors_in_trial}]", request=reply)

        actor_class = self.cog_settings.actor_classes[self_info.actor_class]

        config = None
        if reply.HasField("config"):
            if actor_class.config_type is None:
                raise Exception(
                    f"Actor [{self_info.name}] received config data of unknown type (was it defined in cogment.yaml)")
            config = actor_class.config_type()
            config.ParseFromString(reply.config.content)

        new_session = _ClientActorSession(
            impl, actor_class, trial, self_info.name, impl_name, config, self._actor_stub
        )

        metadata = (("trial-id", trial.id), ("actor-name", self_info.name))
        req_itor = write_actions(new_session)
        reply_itor = self._actor_stub.ActionStream(request_iterator=req_itor, metadata=metadata)

        reader_task = asyncio.create_task(read_observations(new_session, reply_itor))

        try:
            new_session._task = asyncio.create_task(new_session._run())
            await new_session._task
            logging.debug(f"User client implementation for [{new_session.name}] returned")

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise

        finally:
            reader_task.cancel()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/hooks.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='cogment/api/hooks.proto',
  package='cogment',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x17\x63ogment/api/hooks.proto\x12\x07\x63ogment\x1a\x18\x63ogment/api/common.proto\"[\n\x0fPreTrialContext\x12\x11\n\timpl_name\x18\x01 \x01(\t\x12$\n\x06params\x18\x02 \x01(\x0b\x32\x14.cogment.TrialParams\x12\x0f\n\x07user_id\x18\x03 \x01(\t2\x8c\x01\n\nTrialHooks\x12\x42\n\nOnPreTrial\x12\x18.cogment.PreTrialContext\x1a\x18.cogment.PreTrialContext\"\x00\x12:\n\x07Version\x12\x17.cogment.VersionRequest\x1a\x14.cogment.VersionInfo\"\x00\x62\x06proto3'
  ,
  dependencies=[cogment_dot_api_dot_common__pb2.DESCRIPTOR,])




_PRETRIALCONTEXT = _descriptor.Descriptor(
  name='PreTrialContext',
  full_name='cogment.PreTrialContext',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='impl_name', full_name='cogment.PreTrialContext.impl_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='params', full_name='cogment.PreTrialContext.params', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='cogment.PreTrialContext.user_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=62,
  serialized_end=153,
)

_PRETRIALCONTEXT.fields_by_name['params'].message_type = cogment_dot_api_dot_common__pb2._TRIALPARAMS
DESCRIPTOR.message_types_by_name['PreTrialContext'] = _PRETRIALCONTEXT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PreTrialContext = _reflection.GeneratedProtocolMessageType('PreTrialContext', (_message.Message,), {
  'DESCRIPTOR' : _PRETRIALCONTEXT,
  '__module__' : 'cogment.api.hooks_pb2'
  # @@protoc_insertion_point(class_scope:cogment.PreTrialContext)
  })
_sym_db.RegisterMessage(PreTrialContext)



_TRIALHOOKS = _descriptor.ServiceDescriptor(
  name='TrialHooks',
  full_name='cogment.TrialHooks',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=156,
  serialized_end=296,
  methods=[
  _descriptor.MethodDescriptor(
    name='OnPreTrial',
    full_name='cogment.TrialHooks.OnPreTrial',
    index=0,
    containing_service=None,
    input_type=_PRETRIALCONTEXT,
    output_type=_PRETRIALCONTEXT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Version',
    full_name='cogment.TrialHooks.Version',
    index=1,
    containing_service=None,
    input_type=cogment_dot_api_dot_common__pb2._VERSIONREQUEST,
    output_type=cogment_dot_api_dot_common__pb2._VERSIONINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_TRIALHOOKS)

DESCRIPTOR.services_by_name['TrialHooks'] = _TRIALHOOKS

# @@protoc_insertion_point(module_scope)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import SimpleNamespace

import cogment.api.hooks_pb2_grpc as grpc_api
import cogment.api.hooks_pb2 as hooks_api
import cogment.utils as utils
from cogment.trial import Trial
from cogment.prehook import _ServedPrehookSession

import logging
import traceback


class PrehookServicer(grpc_api.TrialHooksServicer):

    def __init__(self, impls, cog_settings, prometheus_registry):

        self.__impls = impls
        self.__cog_settings = cog_settings

        logging.info("Prehook Service started")

    async def OnPreTrial(self, request, context):
        try:
            metadata = dict(context.invocation_metadata())
            trial_id = metadata["trial-id"]

            trial = Trial(trial_id, [], self.__cog_settings)
            user_params = utils.raw_params_to_user_params(request.params, self.__cog_settings)

            prehook = _ServedPrehookSession(user_params, trial)
            for impl in self.__impls:
                try:
                    await impl(prehook)
                except Exception:
                    logging.error(
                        f"An exception occured in user pre-trial hook implementation:\n{traceback.format_exc()}")
                    raise

                prehook._recode()

            reply = hooks_api.PreTrialContext()
            reply.CopyFrom(request)
            reply.params.CopyFrom(utils.user_params_to_raw_params(prehook._params, self.__cog_settings))
            return reply

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import traceback
from abc import ABC, abstractmethod


class DatalogSession(ABC):
    """This represents a datalogger working locally."""

    def __init__(self, impl, trial, trial_params):
        self.trial_id = trial.id
        self.trial_params = trial_params

        self.on_sample = None
        self.on_trial_over = None

        self._started = False
        self._task = None

        self.__impl = impl
        self.__sample_future = None

    def start(self):
        self._started = True

    async def _end(self):
        if self.on_trial_over is not None:
            self.on_trial_over()
        self.__sample_future = None

    async def get_sample(self):
        assert self.on_sample is None

        self.__sample_future = asyncio.get_running_loop().create_future()
        return await self.__sample_future

    async def get_all_samples(self):
        while True:
            sample = await self.get_sample()
            if sample is None:
                break
            yield sample

    def _new_sample(self, sample):
        if self.on_sample is not None:
            self.on_sample(sample)
            self.__sample_future = None
        elif self.__sample_future is not None:
            self.__sample_future.set_result(sample)
            self.__sample_future = None
        else:
            logging.warning("A sample was missed")

    async def _run(self):
        try:
            await self.__impl(self)
        except Exception:
            logging.error(f"An exception occured in user datalog implementation:\n{traceback.format_exc()}")
            raise


class _ServedDatalogSession(DatalogSession):
    def __init__(self, impl, trial, trial_params):
        super().__init__(impl, trial, trial_params)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cogment.api.agent_pb2_grpc as grpc_api
import cogment.api.agent_pb2 as agent_api

from cogment.utils import list_versions
from cogment.trial import Trial

from cogment.errors import InvalidRequestError
from cogment.delta_encoding import DecodeObservationData
from cogment.actor import _ServedActorSession, RecvReward

from prometheus_client import Summary, Counter, Gauge

from types import SimpleNamespace
import traceback
import atexit
import logging
import asyncio
import grpc.experimental.aio


def _trial_key(trial_id, actor_name):
    return f"{trial_id}_{actor_name}"


def _impl_can_serve_actor_class(impl, actor_class):
    if impl.actor_classes:
        for ac in impl.actor_classes:
            if ac == actor_class.id:
                return True
        return False
    else:
        return True


async def read_observations(context, agent_session):
    try:
        while True:
            request = await context.read()

            if request == grpc.experimental.aio.EOF:
                logging.info(f"The orchestrator disconnected agent [{agent_session.name}]")
                break

            agent_session._trial.tick_id = request.observation.tick_id

            obs = DecodeObservationData(
                agent_session._actor_class,
                request.observation.data,
                agent_session._latest_observation,
            )
            agent_session._latest_observation = obs
            agent_session._new_observation(obs)

    except asyncio.CancelledError:
        logging.debug(f"Agent [{agent_session.name}] 'read_observations' coroutine cancelled")

    except Exception:
        logging.error(f"{traceback.format_exc()}")
        raise


async def write_actions(context, agent_session):
    try:
        while True:
            act = await agent_session._retrieve_action()
            msg = agent_api.AgentActionReply()

            if act is not None:
                msg.action.content = act.SerializeToString()

            msg.rewards.extend(agent_session._trial._gather_all_rewards())

            actor_name = dict(context.invocation_metadata())["actor-name"]
            msg.messages.extend(agent_session._trial._gather_all_messages(actor_name))

            await context.write(msg)

    except asyncio.CancelledError:
        logging.debug(f"Agent [{agent_session.name}] 'write_action' coroutine cancelled")

    except Exception:
        logging.error(f"{traceback.format_exc()}")
        raise


class AgentServicer(grpc_api.AgentEndpointServicer):
    def __init__(self, agent_impls, cog_settings, prometheus_registry):
        self.__impls = agent_impls
        self.__agent_sessions = {}
        self.__cog_settings = cog_settings
        atexit.register(self.__cleanup)

        self.decide_request_time = Summary(
            "actor_decide_processing_seconds",
            "Time spent by an actor on the decide function",
            ["name", "actor_class"],
            registry=prometheus_registry
        )
        self.actors_started = Counter(
            "actor_started", "Number of actors created", ["actor_class"],
            registry=prometheus_registry
        )

        self.actors_ended = Counter(
            "actor_ended", "Number of actors ended", ["actor_class"],
            registry=prometheus_registry
        )
        self.messages_received = Counter(
            "actor_received_messages", "Number of messages received", ["name", "actor_class"],
            registry=prometheus_registry
        )
        self.rewards_received = Gauge(
            "actor_reward_summation", "Cumulative rewards received", ["name", "actor_class"],
            registry=prometheus_registry
        )
        self.rewards_counter = Counter(
            "actor_rewards_count", "Number of rewards received", ["name", "actor_class"],
            registry=prometheus_registry
        )

        logging.info("Agent Service started")

    async def OnStart(self, request, context):
        try:
            metadata = dict(context.invocation_metadata())
            actor_name = str(metadata["actor-name"])
            trial_id = metadata["trial-id"]

            if len(self.__impls) == 0:
                raise InvalidRequestError(message="No implementation registered", request=request)

            if request.impl_name:
                impl = self.__impls.get(request.impl_name)
                if impl is None:
                    raise InvalidRequestError(message=f"Unknown agent impl [{request.impl_name}]", request=request)
            else:
                impl_name, impl = next(iter(self.__impls.items()))
                logging.info(f"Agent impl [{impl_name}] arbitrarily chosen for actor [{actor_name}]")

            self_info = None
            for info in request.actors_in_trial:
                if info.name == actor_name:
                    self_info = info
                    break
            if self_info is None:
                logging.debug(f"OnStart [{metadata}] request [{request}]")
                logging.debug(f"trial id [{trial_id}] agent name [{actor_name}]")
                raise InvalidRequestError(f"Unknown agent name [{actor_name}]", request=request)

            if self_info.actor_class not in self.__cog_settings.actor_classes:
                raise InvalidRequestError(
                    message=f"Unknown agent class [{request.actor_class}]", request=request
                )
            actor_class = self.__cog_settings.actor_classes[self_info.actor_class]

            if not _impl_can_serve_actor_class(impl, actor_class):
                raise InvalidRequestError(
                    message=f"[{request.impl_name}] does not implement [{request.actor_class}]",
                    request=request,
                )

            key = _trial_key(trial_id, actor_name)
            if key in self.__agent_sessions:
                raise InvalidRequestError(message="Agent already exists", request=request)

            self.actors_started.labels(request.impl_name).inc()

            trial = Trial(trial_id, request.actors_in_trial, self.__cog_settings)

            config = None
            if request.HasField("config"):
                if actor_class.config_type is None:
                    raise Exception(
                        f"Actor [{actor_name}] received config data of unknown type (was it defined in cogment.yaml?)")
                config = actor_class.config_type()
                config.ParseFromString(request.config.content)

            new_session = _ServedActorSession(
                impl.impl, actor_class, trial, self_info.name, request.impl_name, config
            )
            self.__agent_sessions[key] = new_session

            new_session._task = asyncio.create_task(new_session._run())

            return agent_api.AgentStartReply()

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise

    async def OnEnd(self, request, context):
        try:
            metadata = dict(context.invocation_metadata())
            key = _trial_key(metadata["trial-id"], metadata["actor-name"])

            agent_session = self.__agent_sessions.get(key)
            if agent_session is not None:
                package = SimpleNamespace(observations=[], rewards=[], messages=[])
                for obs_request in request.final_data.observations:
                    obs = DecodeObservationData(
                        agent_session._actor_class,
                        obs_request.data,
                        agent_session._latest_observation)
                    agent_session._latest_observation = obs
                    package.observations.append(obs)

                for rew_request in request.final_data.rewards:
                    reward = RecvReward()
                    reward._set_all(rew_request)
                    package.rewards.append(reward)

                for msg_request in request.final_data.messages:
                    package.messages.append(msg_request)

                await agent_session._end(package)

                self.actors_ended.labels(agent_session.impl_name).inc()

                del self.__agent_sessions[key]

            else:
                logging.error(f"Unknown trial id [{trial_id}] or actor name [{actor_name}] for end-of-trial")

            return agent_api.AgentEndReply()

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise

    async def OnObservation(self, request_iterator, context):
        reader_task = None
        writer_task = None
        try:
            metadata = dict(context.invocation_metadata())
            key = _trial_key(metadata["trial-id"], metadata["actor-name"])

            agent_session = self.__agent_sessions.get(key)
            if agent_session is not None:
                with self.decide_request_time.labels(agent_session.name, agent_session.impl_name).time():
                    reader_task = asyncio.create_task(read_observations(context, agent_session))
                    writer_task = asyncio.create_task(write_actions(context, agent_session))

                    await agent_session._task
                    logging.debug(f"User agent implementation for [{agent_session.name}] returned")

                    if key in self.__agent_sessions:
                        logging.error(f"User agent implementation for [{agent_session.name}] ended before required")
                        del self.__agent_sessions[key]
            else:
                logging.error(f"Unknown trial id [{trial_id}] or actor name [{actor_name}] for observation")

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise

        finally:
            if reader_task is not None:
                reader_task.cancel()
            if writer_task is not None:
                writer_task.cancel()

    async def OnReward(self, request, context):
        try:
            metadata = dict(context.invocation_metadata())
            actor_name = str(metadata["actor-name"])
            trial_id = metadata["trial-id"]

            key = _trial_key(trial_id, actor_name)

            reward = RecvReward()
            reward._set_all(request.reward)

            agent_session = self.__agent_sessions.get(key)
            if agent_session is not None:
                self.rewards_counter.labels(agent_session.name, agent_session.impl_name).inc()
                if reward.value < 0.0:
                    self.rewards_received.labels(agent_session.name, agent_session.impl_name).dec(abs(reward.value))
                else:
                    self.rewards_received.labels(agent_session.name, agent_session.impl_name).inc(reward.value)

                agent_session._new_reward(reward)
            else:
                logging.error(f"Unknown trial id [{trial_id}] or actor name [{actor_name}] for reward")

            return agent_api.AgentRewardReply()

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise

    async def OnMessage(self, request, context):
        try:
            metadata = dict(context.invocation_metadata())
            actor_name = str(metadata["actor-name"])
            trial_id = metadata["trial-id"]

            key = _trial_key(trial_id, actor_name)

            agent_session = self.__agent_sessions.get(key)
            if agent_session is not None:
                for message in request.messages:
                    agent_session._new_message(message)
                    self.messages_received.labels(agent_session.name, agent_session.impl_name).inc()
            else:
                logging.error(f"Unknown trial id [{trial_id}] or actor name [{actor_name}] for message")

            return agent_api.AgentMessageReply()

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise

    async def Version(self, request, context):
        try:
            return list_versions()

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise

    def __cleanup(self):
        self.__agent_sessions.clear()
        atexit.unregister(self.__cleanup)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cogment']

package_data = \
{'': ['*']}

install_requires = \
['grpcio-reflection',
 'grpcio==1.30',
 'prometheus-client>=0.8.0,<0.9.0',
 'protobuf>=3.7']

setup_kwargs = {
    'name': 'cogment',
    'version': '1.0.0a7',
    'description': 'Cogment python SDK',
    'long_description': '# cogment python SDK\n\n> ⚠️ 🚧 This is part of an upcoming release of cogment and still unstable.\n>\n> Current stable version can be found at <https://gitlab.com/cogment/cogment>\n\n[![Latest final release](https://img.shields.io/pypi/v/cogment?style=flat-square)](https://pypi.org/project/cogment/) [![Apache 2 License](https://img.shields.io/badge/license-Apache%202-green?style=flat-square)](./LICENSE) [![Changelog](https://img.shields.io/badge/-Changelog%20-blueviolet?style=flat-square)](./CHANGELOG.md)\n\n## Introduction\n\nThe Cogment framework is a high-efficiency, open source framework designed to enable the training of models in environments where humans and agents interact with the environment and each other continuously. It’s capable of distributed, multi-agent, multi-model training.\n\nThis is the python API for making use of the cogment framework when working with the Python programming language.\n\nFor further Cogment information, check out the documentation at <https://docs.cogment.ai>\n\n## Developers\n\n### Local setup\n\nMake sure you have the following installed:\n  - [Python](https://www.python.org) (any version >=3.7 should work),\n  - [Poetry](https://python-poetry.org).\n\nInstall the dependencies, including downloading and building the cogment protobuf API, by navigating to the python SDK directory and run the following\n\n```\npoetry install\n```\n\n### Define used Cogment protobuf API\n\nThe version of the used cogment protobuf API is defined in the `.cogment-api.yaml` file at the root of the repository. The following can be defined:\n\n- `cogment_api_version: "latest"`, is the default, it retrieves the _latest_ build of the cogment-api `develop`,\n- `cogment_api_version: "vMAJOR.MINOR.PATCH[-PRERELEASE]"`, retrieves an official release of cogment-api.\n- `cogment_api_path: "../path/to/cogment-api"`, retrieves a local version of cogment-api found at the given path ; if set, this overrides `cogment_api_version`.\n\n> ⚠️ when building a docker image, `cogment_api_path` needs to exists in the docker file system. In practice it means it should be a subdirectory of the current directory.\n\n### Tests\n\nTo run them the first step is to configure the way to launch the orchestrator and the cli in a `.env` file. \n\nYou can copy `.env.template` for an example of what\'s expected.\n\n#### Module tests\n\nThese tests only rely on the sdk, no connection to an orchestrator is done.\n\nTo execute the module tests, simply run\n\n```\npoetry run task test\n```\n\n#### Integration tests\n\nThese tests launch and use an orchestrator they are slower but more in depth. \n\nThen, to execute the integration tests (as well as the module tests), simply run\n\n```\npoetry run task test --launch-orchestrator\n```\n\nThese tests can also be launched in a docker image.\n\n```\ndocker build \\\n  -t cogment/cogment-py-sdk-integration-test:latest \\\n  --build-arg COGMENT_IMAGE="<PATH_TO_COGMENT_IMAGE" \\\n  --build-arg COGMENT_ORCHESTRATOR_IMAGE="<PATH_TO_COGMENT_ORCHESTRATOR_IMAGE" \\\n  -f integration_test.dockerfile .\ndocker run --rm cogment/cogment-py-sdk-integration-test:latest\n```\n\n### Lint\n\nRun the linter using\n\n```\npoetry run task lint\n```\n\n### Build the source package\n\nBuild the source package (this step will only be succesfull if `poetry install` succeeded)\n\n```\npoetry build -f sdist\n```\n\n### Build a Docker image\n\nNavigate to the python SDK directory and run the following in order to create an image that can be used by a cogment project:\n\n```\ndocker build -t image_name .\n```\n\n\n',
    'author': 'Artificial Intelligence Redefined',
    'author_email': 'dev+cogment@ai-r.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://cogment.ai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/datalog.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2
from cogment.api import environment_pb2 as cogment_dot_api_dot_environment__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='cogment/api/datalog.proto',
  package='cogment',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x19\x63ogment/api/datalog.proto\x12\x07\x63ogment\x1a\x18\x63ogment/api/common.proto\x1a\x1d\x63ogment/api/environment.proto\"y\n\x18LogExporterSampleRequest\x12,\n\x0ctrial_params\x18\x01 \x01(\x0b\x32\x14.cogment.TrialParamsH\x00\x12(\n\x06sample\x18\x02 \x01(\x0b\x32\x16.cogment.DatalogSampleH\x00\x42\x05\n\x03msg\"\xb7\x01\n\rDatalogSample\x12\x0f\n\x07user_id\x18\x01 \x01(\t\x12-\n\x0cobservations\x18\x02 \x01(\x0b\x32\x17.cogment.ObservationSet\x12 \n\x07\x61\x63tions\x18\x03 \x03(\x0b\x32\x0f.cogment.Action\x12 \n\x07rewards\x18\x04 \x03(\x0b\x32\x0f.cogment.Reward\x12\"\n\x08messages\x18\x05 \x03(\x0b\x32\x10.cogment.Message\"\x18\n\x16LogExporterSampleReply2\xa0\x01\n\x0bLogExporter\x12U\n\x0bOnLogSample\x12!.cogment.LogExporterSampleRequest\x1a\x1f.cogment.LogExporterSampleReply\"\x00(\x01\x12:\n\x07Version\x12\x17.cogment.VersionRequest\x1a\x14.cogment.VersionInfo\"\x00\x62\x06proto3'
  ,
  dependencies=[cogment_dot_api_dot_common__pb2.DESCRIPTOR,cogment_dot_api_dot_environment__pb2.DESCRIPTOR,])




_LOGEXPORTERSAMPLEREQUEST = _descriptor.Descriptor(
  name='LogExporterSampleRequest',
  full_name='cogment.LogExporterSampleRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial_params', full_name='cogment.LogExporterSampleRequest.trial_params', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample', full_name='cogment.LogExporterSampleRequest.sample', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='msg', full_name='cogment.LogExporterSampleRequest.msg',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=95,
  serialized_end=216,
)


_DATALOGSAMPLE = _descriptor.Descriptor(
  name='DatalogSample',
  full_name='cogment.DatalogSample',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_id', full_name='cogment.DatalogSample.user_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='observations', full_name='cogment.DatalogSample.observations', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actions', full_name='cogment.DatalogSample.actions', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rewards', full_name='cogment.DatalogSample.rewards', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='messages', full_name='cogment.DatalogSample.messages', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=219,
  serialized_end=402,
)


_LOGEXPORTERSAMPLEREPLY = _descriptor.Descriptor(
  name='LogExporterSampleReply',
  full_name='cogment.LogExporterSampleReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=404,
  serialized_end=428,
)

_LOGEXPORTERSAMPLEREQUEST.fields_by_name['trial_params'].message_type = cogment_dot_api_dot_common__pb2._TRIALPARAMS
_LOGEXPORTERSAMPLEREQUEST.fields_by_name['sample'].message_type = _DATALOGSAMPLE
_LOGEXPORTERSAMPLEREQUEST.oneofs_by_name['msg'].fields.append(
  _LOGEXPORTERSAMPLEREQUEST.fields_by_name['trial_params'])
_LOGEXPORTERSAMPLEREQUEST.fields_by_name['trial_params'].containing_oneof = _LOGEXPORTERSAMPLEREQUEST.oneofs_by_name['msg']
_LOGEXPORTERSAMPLEREQUEST.oneofs_by_name['msg'].fields.append(
  _LOGEXPORTERSAMPLEREQUEST.fields_by_name['sample'])
_LOGEXPORTERSAMPLEREQUEST.fields_by_name['sample'].containing_oneof = _LOGEXPORTERSAMPLEREQUEST.oneofs_by_name['msg']
_DATALOGSAMPLE.fields_by_name['observations'].message_type = cogment_dot_api_dot_environment__pb2._OBSERVATIONSET
_DATALOGSAMPLE.fields_by_name['actions'].message_type = cogment_dot_api_dot_common__pb2._ACTION
_DATALOGSAMPLE.fields_by_name['rewards'].message_type = cogment_dot_api_dot_common__pb2._REWARD
_DATALOGSAMPLE.fields_by_name['messages'].message_type = cogment_dot_api_dot_common__pb2._MESSAGE
DESCRIPTOR.message_types_by_name['LogExporterSampleRequest'] = _LOGEXPORTERSAMPLEREQUEST
DESCRIPTOR.message_types_by_name['DatalogSample'] = _DATALOGSAMPLE
DESCRIPTOR.message_types_by_name['LogExporterSampleReply'] = _LOGEXPORTERSAMPLEREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LogExporterSampleRequest = _reflection.GeneratedProtocolMessageType('LogExporterSampleRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGEXPORTERSAMPLEREQUEST,
  '__module__' : 'cogment.api.datalog_pb2'
  # @@protoc_insertion_point(class_scope:cogment.LogExporterSampleRequest)
  })
_sym_db.RegisterMessage(LogExporterSampleRequest)

DatalogSample = _reflection.GeneratedProtocolMessageType('DatalogSample', (_message.Message,), {
  'DESCRIPTOR' : _DATALOGSAMPLE,
  '__module__' : 'cogment.api.datalog_pb2'
  # @@protoc_insertion_point(class_scope:cogment.DatalogSample)
  })
_sym_db.RegisterMessage(DatalogSample)

LogExporterSampleReply = _reflection.GeneratedProtocolMessageType('LogExporterSampleReply', (_message.Message,), {
  'DESCRIPTOR' : _LOGEXPORTERSAMPLEREPLY,
  '__module__' : 'cogment.api.datalog_pb2'
  # @@protoc_insertion_point(class_scope:cogment.LogExporterSampleReply)
  })
_sym_db.RegisterMessage(LogExporterSampleReply)



_LOGEXPORTER = _descriptor.ServiceDescriptor(
  name='LogExporter',
  full_name='cogment.LogExporter',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=431,
  serialized_end=591,
  methods=[
  _descriptor.MethodDescriptor(
    name='OnLogSample',
    full_name='cogment.LogExporter.OnLogSample',
    index=0,
    containing_service=None,
    input_type=_LOGEXPORTERSAMPLEREQUEST,
    output_type=_LOGEXPORTERSAMPLEREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Version',
    full_name='cogment.LogExporter.Version',
    index=1,
    containing_service=None,
    input_type=cogment_dot_api_dot_common__pb2._VERSIONREQUEST,
    output_type=cogment_dot_api_dot_common__pb2._VERSIONINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_LOGEXPORTER)

DESCRIPTOR.services_by_name['LogExporter'] = _LOGEXPORTER

# @@protoc_insertion_point(module_scope)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import importlib
import logging
import traceback
from cogment.session import Session
from abc import ABC

ENVIRONMENT_ACTOR_NAME = "env"

_ACTIONS = "actions"
_FINAL_ACTIONS = "final_actions"
_MESSAGE = "message"


class EnvironmentSession(Session):
    """This represents the environment being performed locally."""

    def __init__(self, impl, trial, impl_name, config):
        super().__init__(trial)
        self.impl_name = impl_name
        self.config = config

        self._ended = False
        self._obs_queue = asyncio.Queue()
        self._last_event_received = False
        self._task = None  # Task used to call _run()

        self.__impl = impl
        self.__started = False
        self.__event_queue = None
        self.__final_obs_future = None

    def start(self, observations):
        assert not self.__started
        assert not self._ended

        self.__event_queue = asyncio.Queue()
        self.__started = True

        self._obs_queue.put_nowait((observations, False))

    async def event_loop(self):
        assert self.__started
        assert not self._ended

        loop_active = not self._last_event_received
        while loop_active:
            try:
                event = await self.__event_queue.get()

            except asyncio.CancelledError:
                logging.debug("Coroutine cancelled while waiting for an event")
                break

            self._last_event_received = _FINAL_ACTIONS in event
            keep_looping = yield event
            self.__event_queue.task_done()
            loop_active = ((keep_looping is None or bool(keep_looping)) and
                            not self._last_event_received and
                            not self._ended)

            if not loop_active:
                if self._last_event_received:
                    logging.debug(f"Last event received, exiting environment event loop")
                elif self._ended:
                    logging.debug(f"Last observation sent, exiting environment event loop")
                else:
                    logging.debug(f"End of event loop for environment requested by user")

        logging.debug(f"Exiting environment event loop generator")

    def produce_observations(self, observations):
        assert self.__started
        self._obs_queue.put_nowait((observations, False))

    def end(self, observations):
        if self.__final_obs_future is not None:
            self.__final_obs_future.set_result(observations)
        elif not self._ended:
            self._ended = True
            self._obs_queue.put_nowait((observations, True))

    async def _retrieve_obs(self):
        obs = await self._obs_queue.get()
        self._obs_queue.task_done()
        return obs

    async def _end_request(self, actions):
        logging.debug("Environment received an end request")

        if self._ended:
            return None
        self._ended = True
        if not self.__started:
            return None

        result = None
        if self.__event_queue:
            self.__final_obs_future = asyncio.get_running_loop().create_future()

            event = {_FINAL_ACTIONS : actions}
            await self.__event_queue.put(event)

            result = await self.__final_obs_future
            self.__final_obs_future = None
        else:
            logging.warning("The environment received an end request that it was unable to handle.")

        return result

    def _new_action(self, actions):
        if not self.__started or self._ended:
            return

        if self.__event_queue:
            event = {_ACTIONS : actions}
            self.__event_queue.put_nowait(event)
        else:
            logging.warning("The environment received actions that it was unable to handle.")

    def _new_message(self, message):
        logging.debug("Environment received a message")

        if not self.__started or self._ended:
            return

        if self.__event_queue:
            event = {_MESSAGE : (message.sender_name(), message.payload())}
            self.__event_queue.put_nowait(event)
        else:
            logging.warning("The environment received a message that it was unable to handle.")

    async def _run(self):
        try:
            await self.__impl(self)

        except asyncio.CancelledError:
            logging.debug("Environment implementation coroutine cancelled")

        except Exception:
            logging.error(f"An exception occured in user environment implementation:\n{traceback.format_exc()}")
            raise


class _ServedEnvironmentSession(EnvironmentSession):
    """An environment session that is served from an environment service."""

    def __init__(self, impl, trial, impl_name, config):
        super().__init__(impl, trial, impl_name, config)

    def add_reward(self, value, confidence, to, tick_id=-1, user_data=None):
        assert self._trial is not None
        self._trial.add_reward(value, confidence, to, tick_id, user_data)

    def send_message(self, user_data, to, to_environment=False):
        assert self._trial is not None
        self._trial.send_message(user_data, to, to_environment)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cogment.api.datalog_pb2_grpc as grpc_api
import cogment.api.datalog_pb2 as datalog_api
from cogment.trial import Trial
from cogment.datalog import _ServedDatalogSession
from cogment.utils import raw_params_to_user_params, list_versions
import logging
import traceback
import asyncio


class LogExporterService(grpc_api.LogExporterServicer):

    def __init__(self, impl, cog_settings):
        self.__impl = impl
        self.__cog_settings = cog_settings
        logging.info("Log Exporter Service started")

    async def OnLogSample(self, request_iterator, context):
        try:
            metadata = dict(context.invocation_metadata())
            trial_id = metadata["trial-id"]

            trial = Trial(trial_id, [], self.__cog_settings)

            msg = await request_iterator.__anext__()
            assert msg.HasField("trial_params")
            trial_params = raw_params_to_user_params(msg.trial_params, self.__cog_settings)

            session = _ServedDatalogSession(self.__impl, trial, trial_params)
            loop = asyncio.get_running_loop()
            session._task = loop.create_task(session._run())

            # TODO: Wait for session._started
            async for msg in request_iterator:
                if session._task.done():
                    break
                assert msg.HasField("sample")
                session._new_sample(msg.sample)

            if not session._task.done():
                session._end()

            return datalog_api.LogExporterSampleReply()

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise

    async def Version(self, request, context):
        try:
            return list_versions()
        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise
